--imports
local assert = require("justo.assert")
local justo = require("justo")
local suite, test, init = justo.suite, justo.test, justo.init
local Trans = require("dogma.trans.js.Trans")
local Parser = require("dogma.syn.Parser")

--Suite.
return suite("dogma.trans.js._.StmtTrans", function()
  local trans, parser

  ----------
  -- init --
  ----------
  init("*", function()
    parser = Parser.new()
    trans = Trans.new()
    trans:transform(parser)
  end):title("Create transformer")

  ---------
  -- use --
  ---------
  suite("use", function()
    test("use LiteralStr", function()
      parser:parse('use "my/module"')
      assert(trans:next()):eq('const module = dogma.use(require("my/module"));\n')
    end)

    test("use LiteralStr as Name", function()
      parser:parse('use "my/module" as mod')
      assert(trans:next()):eq('const mod = dogma.use(require("my/module"));\n')
    end)

    test("use LiteralStr, LiteralStr", function()
      parser:parse('use "my/mod1", "my/mod2"')
      assert(trans:next()):eq('const mod1 = dogma.use(require("my/mod1"));const mod2 = dogma.use(require("my/mod2"));\n')
    end)

    test("use LiteralStr as Name, LiteralStr as Name", function()
      parser:parse('use "my/module1" as mod1, "my/module2" as mod2')
      assert(trans:next()):eq('const mod1 = dogma.use(require("my/module1"));const mod2 = dogma.use(require("my/module2"));\n')
    end)

    test("use module as Name => {variables}", function()
      parser:parse('use "my/mod1" => {var1,var2}, "my/mod2" => {var3,var4}')
      assert(trans:next()):eq('const mod1 = dogma.use(require("my/mod1"));let {var1,var2} = mod1;const mod2 = dogma.use(require("my/mod2"));let {var3,var4} = mod2;\n')
    end)

    test("use <alias>", function()
      parser:parse("use <fs>", nil, {imports={fs="@dogmalang/fs2"}})
      assert(trans:next()):eq('const fs = dogma.use(require("@dogmalang/fs2"));\n')
    end)

    test("use <alias> as other", function()
      parser:parse("use <fs> as fs1", nil, {imports={fs="@dogmalang/fs"}})
      assert(trans:next()):eq('const fs1 = dogma.use(require("@dogmalang/fs"));\n')
    end)
  end):tags("use")

  ----------
  -- from --
  ----------
  suite("from", function()
    test("from LiteralStr use Name", function()
      parser:parse([[from "fs" use FSWatcher]])
      assert(trans:next()):eq('import {FSWatcher} from "fs";\n')
    end)

    test("from LiteralStr use Name as Name", function()
      parser:parse([[from "fs" use FSWatcher as Watcher]])
      assert(trans:next()):eq('import {FSWatcher as Watcher} from "fs";\n')
    end)

    test("from LiteralStr use Name, Name", function()
      parser:parse([[from "fs" use FSWatcher, watch]])
      assert(trans:next()):eq('import {FSWatcher, watch} from "fs";\n')
    end)

    test("from LiteralStr use Name as Name, Name as Name", function()
      parser:parse([[from "fs" use FSWatcher as Watcher, watch as w]])
      assert(trans:next()):eq('import {FSWatcher as Watcher, watch as w} from "fs";\n')
    end)

    test("from module use Name=>{variables}", function()
      parser:parse([[from "module" use mem1 as MEM1=>{var1,var2}, mem2=>{var3,var4}]])
      assert(trans:next()):eq('import {mem1 as MEM1, mem2} from "module";let {var1,var2} = MEM1;let {var3,var4} = mem2;\n')
    end)
  end)

  -----------
  -- break --
  -----------
  suite("break", function()
    test("break", function()
      parser:parse("break")
      assert(trans:next()):eq("break;\n")
    end)

    test("break for", function(params)
      parser:parse("break " .. params[1])
      assert(trans:next()):eq("$__break__ = true; break;\n")
    end):iter(
      {subtitle=" # for", params="for"},
      {subtitle=" # for each", params="for each"},
      {subtitle=" # while", params="while"}
    )
  end)

  ----------
  -- next --
  ----------
  test("next", function()
    parser:parse("next")
    assert(trans:next()):eq("continue;\n")
  end)

  ----------
  -- enum --
  ----------
  suite("enum", function()
    test("export enum Name {Name}", function()
      parser:parse("export enum Color {RED}")
      assert(trans:next()):eq([[
class Color {
  constructor(name, val) {
    Object.defineProperty(this, "name", {value: name, enum: true});
    Object.defineProperty(this, "value", {value: val, enum: true});
  }

  /* istanbul ignore next */
  toString() {
    return this.name;
  }
}
module.exports = exports = Color;Object.defineProperty(Color, "RED", {value: new Color("RED", 1), enum: true});

]])
    end)

    test("pub enum Name {Name}", function()
      parser:parse("pub enum Color {RED}")
      assert(trans:next()):eq([[
class Color {
  constructor(name, val) {
    Object.defineProperty(this, "name", {value: name, enum: true});
    Object.defineProperty(this, "value", {value: val, enum: true});
  }

  /* istanbul ignore next */
  toString() {
    return this.name;
  }
}
export {Color};Object.defineProperty(Color, "RED", {value: new Color("RED", 1), enum: true});

]])
    end)

    test("enum Name {Name, Name, Name}", function()
      parser:parse("enum Color {RED, GREEN, BLUE}")
      assert(trans:next()):eq([[
class Color {
  constructor(name, val) {
    Object.defineProperty(this, "name", {value: name, enum: true});
    Object.defineProperty(this, "value", {value: val, enum: true});
  }

  /* istanbul ignore next */
  toString() {
    return this.name;
  }
}
Object.defineProperty(Color, "RED", {value: new Color("RED", 1), enum: true});
Object.defineProperty(Color, "GREEN", {value: new Color("GREEN", 2), enum: true});
Object.defineProperty(Color, "BLUE", {value: new Color("BLUE", 3), enum: true});

]])
    end)

    test([[enum Name {Name = Str, Name = LiteralStr, Name = LiteralStr}]], function()
      parser:parse([[enum Color {RED = "1", GREEN = "2", BLUE = "3"}]])
      assert(trans:next()):eq([[
class Color {
  constructor(name, val) {
    Object.defineProperty(this, "name", {value: name, enum: true});
    Object.defineProperty(this, "value", {value: val, enum: true});
  }

  /* istanbul ignore next */
  toString() {
    return this.name;
  }
}
Object.defineProperty(Color, "RED", {value: new Color("RED", "1"), enum: true});
Object.defineProperty(Color, "GREEN", {value: new Color("GREEN", "2"), enum: true});
Object.defineProperty(Color, "BLUE", {value: new Color("BLUE", "3"), enum: true});

]])
    end)
  end):tags("enum")

  -----------
  -- async --
  -----------
  suite("async", function()
    test("async without catch", function()
      parser:parse("async\n 1+2\n 3+4")
      assert(trans:next()):eq("setImmediate(async () => {(1+2);(3+4);});\n")
    end)

    test("async with catch without var", function()
      parser:parse("async\n 1+2\n 3+4\ncatch\n 5+6")
      assert(trans:next()):eq("setImmediate(async () => {try {(1+2);(3+4);} catch(_) {(5+6);}});\n")
    end)

    test("async with catch with var", function()
      parser:parse("async\n 1+2\n 3+4\ncatch e\n 5+6")
      assert(trans:next()):eq("setImmediate(async () => {try {(1+2);(3+4);} catch(e) {(5+6);}});\n")
    end)

    test("async with {delay=exp}", function()
      parser:parse("async with {delay = 123*456} 789")
      assert(trans:next()):eq("dogma.asyncly(async () => {789;}, (123*456));\n")
    end)

    test("async with {delay}", function()
      parser:parse("async with {delay} 789")
      assert(trans:next()):eq("dogma.asyncly(async () => {789;}, delay);\n")
    end)

    test("async with {repeat=exp}", function()
      parser:parse("async with {repeat=true} 789")
      assert(trans:next()):eq("dogma.repeatAsyncly(async () => {789;}, {delay: null, repeat: true});\n")
    end)

    test("async with {repeat}", function()
      parser:parse("async with {repeat} 789")
      assert(trans:next()):eq("dogma.repeatAsyncly(async () => {789;}, {delay: null, repeat: repeat});\n")
    end)

    test("async with {delay=exp, repeat=exp}", function()
      parser:parse("async with {delay=10, repeat=5} 789")
      assert(trans:next()):eq("dogma.repeatAsyncly(async () => {789;}, {delay: 10, repeat: 5});\n")
    end)
  end):tags("async")

  ---------
  -- var --
  ---------
  suite("var", function()
    suite("std", function()
      test("export var Name = Exp", function()
        parser:parse("export var x = 3")
        assert(trans:next()):eq("var x = 3;module.exports = exports = x;\n")
      end)

      test("pub var Name = Exp", function()
        parser:parse("pub var x = 3")
        assert(trans:next()):eq("var x = 3;export {x};\n")
      end)

      test("var Name", function()
        parser:parse("var x")
        assert(trans:next()):eq("let x;\n")
      end)

      test("var Name, Name", function()
        parser:parse("var x, y")
        assert(trans:next()):eq("let x;let y;\n")
      end)

      test("var Name = Exp", function()
        parser:parse("var x = 1+2")
        assert(trans:next()):eq("let x = (1+2);\n")
      end)

      test("var Name = Exp, Name = Exp", function()
        parser:parse("var x = 1+2, y = 3+4")
        assert(trans:next()):eq("let x = (1+2);let y = (3+4);\n")
      end)

      test("var Name = Exp, Name ::= Exp", function()
        parser:parse("var x = 1+2, y ::= 3+4")
        assert(trans:next()):eq("let x = (1+2);const y = (3+4);\n")
      end)
    end)

    suite("map", function()
      test("var {Name} = Exp", function()
        parser:parse("var {x} = 1+2")
        assert(trans:next()):eq("let {x} = (1+2);\n")
      end)

      test("pub var {Name} = Exp", function()
        parser:parse("pub var {x} = 1+2")
        assert(trans:next()):eq("var {x} = (1+2);export {x};\n")
      end)

      test("var {Name, Name} = Exp", function()
        parser:parse("var {x, y} = 1+2")
        assert(trans:next()):eq("let {x, y} = (1+2);\n")
      end)

      test("pub var {Name, Name} = Exp", function()
        parser:parse("pub var {x, y} = 1+2")
        assert(trans:next()):eq("var {x, y} = (1+2);export {x, y};\n")
      end)
    end)

    suite("list", function()
      test("var [Name] = Exp", function()
        parser:parse("var [x] = 1+2")
        assert(trans:next()):eq("let [x] = dogma.getArrayToUnpack((1+2), 1);\n")
      end)

      test("var [Name] ::= Exp", function()
        parser:parse("var [x] ::= 1+2")
        assert(trans:next()):eq("const [x] = dogma.getArrayToUnpack((1+2), 1);\n")
      end)

      test("var [Name, Name] = Exp", function()
        parser:parse("var [x, y] = 1+2")
        assert(trans:next()):eq("let [x, y] = dogma.getArrayToUnpack((1+2), 2);\n")
      end)

      test("var [_, Name] = Exp", function()
        parser:parse("var [_, y] = 1+2")
        assert(trans:next()):eq("let [, y] = dogma.getArrayToUnpack((1+2), 2);\n")
      end)

      test("var [Name, _, Name] = Exp", function()
        parser:parse("var [x, _,z] = 1+2")
        assert(trans:next()):eq("let [x, , z] = dogma.getArrayToUnpack((1+2), 3);\n")
      end)

      test("pub var [Name] = Exp", function()
        parser:parse("pub var [x] = 1+2")
        assert(trans:next()):eq("var [x] = dogma.getArrayToUnpack((1+2), 1);export {x};\n")
      end)

      test("var [Name, ...Name] = Exp", function()
        parser:parse("var [x, ...y] = 1+2")
        assert(trans:next()):eq("let [x, ...y] = dogma.getArrayToUnpack((1+2), 2);\n")
      end)

      test("pub var [Name, ...Name] = Exp", function()
        parser:parse("pub var [x, ...y] = 1+2")
        assert(trans:next()):eq("var [x, ...y] = dogma.getArrayToUnpack((1+2), 2);export {x, y};\n")
      end)
    end)
  end):tags("var")

  -----------
  -- const --
  -----------
  suite("const", function()
    test("const Name = Exp", function()
      parser:parse("const x = 1+2")
      assert(trans:next()):eq("const x = (1+2);\n")
    end)

    test("const Name = Exp, Name = Exp", function()
      parser:parse("const x = 1+2, y = 3+4")
      assert(trans:next()):eq("const x = (1+2);const y = (3+4);\n")
    end)

    test("export const Name = Exp", function()
      parser:parse("export const x = 3")
      assert(trans:next()):eq("const x = 3;module.exports = exports = x;\n")
    end)

    test("pub const Name = Exp", function()
      parser:parse("pub const x = 3")
      assert(trans:next()):eq("const x = 3;export {x};\n")
    end)

    test("const {Name} = Exp", function()
      parser:parse("const {x} = 1+2")
      assert(trans:next()):eq("const {x} = (1+2);\n")
    end)

    test("const [Name] = Exp", function()
      parser:parse("const [x] = 1+2")
      assert(trans:next()):eq("const [x] = dogma.getArrayToUnpack((1+2), 1);\n")
    end)
  end):tags("var")

  -----------
  -- while --
  -----------
  suite("while", function()
    test("while Exp do Exp", function()
      parser:parse("while true do x+1")
      assert(trans:next()):eq("while (true) {(x+1);}\n")
    end)

    test("while Exp do Stmt", function()
      parser:parse("while true do return x+1")
      assert(trans:next()):eq("while (true) {return (x+1);}\n")
    end)

    test("while Exp do Block else Block", function()
      parser:parse("while true do x+1\nelse x+2")
      assert(trans:next()):eq("{let $__break__; while (true) {(x+1);} if (!$__break__) {(x+2);}}\n")
    end)

    test("while Exp; Exp do Sent", function()
      parser:parse("while true; x + 1 do x + 2")
      assert(trans:next()):eq("for (; true; (x+1)) {(x+2);}\n")
    end)

    test("while Exp; Exp do Block else Block", function()
      parser:parse("while true; x + 1 do x + 2\nelse x + 3")
      assert(trans:next()):eq("{let $__break__; for (; true; (x+1)) {(x+2);} if (!$__break__) {(x+3);}}\n")
    end)

    test("while with catch", function()
      parser:parse("while true do\n x+1\ncatch\n x+2")
      assert(trans:next()):eq("while (true) try {(x+1);} catch(_) {(x+2);}\n")
    end)

    test("while Exp do Block else Block catch Block", function()
      parser:parse("while true do x+1\nelse x+2\ncatch\n x+3")
      assert(trans:next()):eq("try {let $__break__; while (true) {(x+1);} if (!$__break__) {(x+2);}} catch(_) {(x+3);}\n")
    end)

    test("while with finally", function()
      parser:parse("while true do\n x+1\nfinally\n x+2")
      assert(trans:next()):eq("while (true) try {(x+1);} finally {(x+2);}\n")
    end)

    test("while with catch and finally", function()
      parser:parse("while true do\n x+1\ncatch\n x+2\nfinally\n x+3")
      assert(trans:next()):eq("while (true) try {(x+1);} catch(_) {(x+2);} finally {(x+3);}\n")
    end)
  end):tags("while")

  --------
  -- do --
  --------
  suite("do", function()
    test("do with var", function()
      parser:parse("do var x, y\n x+=1\nwhile true")
      assert(trans:next()):eq("{let x, y;do {(x+=1);} while (true);}\n")
    end)

    test("do with while", function()
      parser:parse("do\n x+1\nwhile true")
      assert(trans:next()):eq("do {(x+1);} while (true);\n")
    end)

    test("do with catch", function()
      parser:parse("do\n x+1\ncatch\n x+2")
      assert(trans:next()):eq("try {(x+1);} catch(_) {(x+2);}\n")
    end)

    test("do with finally", function()
      parser:parse("do\n x+1\nfinally\n x+2")
      assert(trans:next()):eq("try {(x+1);} finally {(x+2);}\n")
    end)

    test("do with catch and finally", function()
      parser:parse("do\n x+1\ncatch\n x+2\nfinally\n x+3")
        assert(trans:next()):eq("try {(x+1);} catch(_) {(x+2);} finally {(x+3);}\n")
    end)

    test("do with while, catch and finally", function()
      parser:parse("do\n x+1\nwhile true\ncatch\n x+2\nfinally\n x+3")
        assert(trans:next()):eq("do try {(x+1);} catch(_) {(x+2);} finally {(x+3);} while (true);\n")
    end)
  end)

  ---------
  -- for --
  ---------
  suite("for", function()
    test("for Name ; Exp do Exp", function()
      parser:parse("for i ; i < 10 do print(i)")
      assert(trans:next()):eq("for (let i; (i<10); ) {print(i);}\n")
    end)

    test("for Name = Exp; Exp do Exp", function()
      parser:parse("for i = 0; i < 10 do print(i)")
      assert(trans:next()):eq("for (let i = 0; (i<10); ) {print(i);}\n")
    end)

    test("for Variable; Exp; Exp do Exp", function()
      parser:parse("for i = 0; i < 10; i += 1 do print(i)")
      assert(trans:next()):eq("for (let i = 0; (i<10); (i+=1)) {print(i);}\n")
    end)

    test("for Name, Name ; Exp do Exp", function()
      parser:parse("for i, j ; i < 10 do print(i)")
      assert(trans:next()):eq("for (let i, j; (i<10); ) {print(i);}\n")
    end)

    test("for Name = Exp, Name = Exp ; Exp do Exp", function()
      parser:parse("for i = 0, j = 1 ; i < 10 do print(i)")
      assert(trans:next()):eq("for (let i = 0, j = 1; (i<10); ) {print(i);}\n")
    end)

    test("for Name = Exp, Name = Exp ; Exp; Exp do Exp", function()
      parser:parse("for i = 0, j = 1 ; i < 10; i += 1 do print(i)")
      assert(trans:next()):eq("for (let i = 0, j = 1; (i<10); (i+=1)) {print(i);}\n")
    end)

    test("for Name = Exp, Name = Exp ; Exp; Exp do Exp else Exp", function()
      parser:parse("for i = 0; i < 10; i += 1 do\n  print(i)\n  if i == 5 then break for\nelse\n  print(j)")
      assert(trans:next()):eq("{let $__break__; for (let i = 0; (i<10); (i+=1)) {print(i);if ((i==5)) {$__break__ = true; break;}} if (!$__break__) {print(j);}}\n")
    end)

    test("for Name; Exp; Exp do\\n Body\\ncatch\\n Body", function()
      parser:parse("for i; i < 10; i += 1 do\n print(i)\ncatch e\n print(e)")
      assert(trans:next()):eq("for (let i; (i<10); (i+=1)) try {print(i);} catch(e) {print(e);}\n")
    end)

    test("for Name; Exp; Exp do Block else Block catch Body", function()
      parser:parse("for i = 0; i < 10; i += 1 do\n print(i)\n  if i == 5 then break for\nelse\n  print(j)\ncatch e\n print(e)")
      assert(trans:next()):eq("try {let $__break__; for (let i = 0; (i<10); (i+=1)) {print(i);if ((i==5)) {$__break__ = true; break;}} if (!$__break__) {print(j);}} catch(e) {print(e);}\n")
    end)
  end):tags("for")

  --------------
  -- for each --
  --------------
  suite("for each", function()
    test("for each Name in Exp do\\n Sent", function(params)
      parser:parse("for each k " .. params[1] .. " arr do\n x+1")
      assert(trans:next()):eq("for (let k of arr) {(x+1);}\n")
    end):iter(
      {subtitle=" # in", params="in"},
      {subtitle=" # :=", params=":="}
    )

    test("for each Name ::= Exp do\\n Sent", function()
      parser:parse("for each k ::= arr do\n x+1")
      assert(trans:next()):eq("for (const k of arr) {(x+1);}\n")
    end)

    test("for each Name ::= Exp do Block else Block", function()
      parser:parse("for each k ::= arr do\n x+1\nelse\n x+2")
      assert(trans:next()):eq("{let $__break__; for (const k of arr) {(x+1);} if (!$__break__) {(x+2);}}\n")
    end)

    test("for each Name, Name in Exp do\\n Sent", function()
      parser:parse("for each k, v in map do\n x+1")
      assert(trans:next()):eq("for (let [k, v] of Object.entries(map)) { {(x+1);} }\n")
    end)

    test("for each Name, Name in Exp do Block else Block", function()
      parser:parse("for each k, v in map do\n x+1\nelse\n x+2")
      assert(trans:next()):eq("{let $__break__; for (let [k, v] of Object.entries(map)) { {(x+1);} if (!$__break__) {(x+2);}}\n")
    end)

    test("for each with catch", function()
      parser:parse("for each k in arr do\n x+1\ncatch\n x+2")
      assert(trans:next()):eq("for (let k of arr) try {(x+1);} catch(_) {(x+2);}\n")
    end)

    test("for each with else and catch", function()
      parser:parse("for each k in arr do\n x+1\nelse\n x+2\ncatch\n x+3")
      assert(trans:next()):eq("try {let $__break__; for (let k of arr) {(x+1);} if (!$__break__) {(x+2);}} catch(_) {(x+3);}\n")
    end)

    test("for each with finally", function()
      parser:parse("for each k in arr do\n x+1\nfinally\n x+2")
      assert(trans:next()):eq("for (let k of arr) try {(x+1);} finally {(x+2);}\n")
    end)

    test("for each with catch and finally", function()
      parser:parse("for each k in arr do\n x+1\ncatch\n x+2\nfinally\n x+3")
      assert(trans:next()):eq("for (let k of arr) try {(x+1);} catch(_) {(x+2);} finally {(x+3);}\n")
    end)
  end):tags("foreach")

  ------------
  -- return --
  ------------
  suite("return", function()
    test("return", function()
      parser:parse("return")
      assert(trans:next()):eq("return;\n")
    end)

    test("return Exp", function()
      parser:parse("return x+1")
      assert(trans:next()):eq("return (x+1);\n")
    end)
  end)

  -----------
  -- yield --
  -----------
  suite("yield", function()
    test("yield Exp", function()
      parser:parse("yield x+1")
      assert(trans:next()):eq("yield (x+1);\n")
    end)
  end):tags("yield")

  ------------
  -- struct --
  ------------
  suite("struct", function()
    suite("template", function()
      test("newline expected", function()
        parser:parse("use Tmpl<T>\nexport struct MyStruct : Tmpl<MyT>var x = 123", "test/data/template/MyStruct.dog")
        assert(function() trans:next() end):raises("test/data/template/MyStruct.dog (2, 35): newline expected.")
      end)

      test("template file not found", function()
        parser:parse("use Tmpl<T>\nexport struct MyStruct : tmpl<MyT>", "test/data/template/MyStruct.dog")
        assert(function() trans:next() end):raises("test/data/template/MyStruct.dog (end of file): file not found: test/data/template/tmpl.dogt.")
      end)

      test("OK", function(params)
        parser:parse(
          string.format("use Tmpl<T>\n%s\nexport struct MyStruct : Tmpl<MyT>\nvar x = 123", params[1]),
          "test/data/template/MyStruct.dog"
        )

        assert(trans:next()):eq([[

const $MyStruct = class MyStruct {
  constructor(_) {
    /* istanbul ignore next */ if (_ == null) _ = {};
    expect('x', _['x'], null);
    Object.defineProperty(this, 'x', {value: coalesce(_['x'], null), writable: true, enumerable: true});
    expect('y', _['y'], null);
    Object.defineProperty(this, 'y', {value: coalesce(_['y'], null), writable: true, enumerable: true});
    /* istanbul ignore next */ if (this._pvt_1858abefb5a2e787ec30ad1f05aa854c___post__ instanceof Function) this._pvt_1858abefb5a2e787ec30ad1f05aa854c___post__();
  }
}
const MyStruct = new Proxy($MyStruct, { apply(receiver, self, args) { return new $MyStruct(...args); } });
module.exports = exports = MyStruct;
const Self = MyStruct;
MyStruct.prototype.inc = function(x) { /* istanbul ignore next */ dogma.paramExpected("x", x, MyT);{} };
MyStruct.prototype.dec = function(x) { /* istanbul ignore next */ dogma.paramExpected("x", x, MyT);{} };
let x = 123;
]])
      end):iter(
        {subtitle=" # with @Self", params="@Self"},
        {subtitle=" # with no @Self", params=""}
      )
    end)

    test("struct Name", function()
      parser:parse("struct MyStruct")
      assert(trans:next()):eq([[
const $MyStruct = class MyStruct {
  constructor(_) {
    /* istanbul ignore next */ if (_ == null) _ = {};
    /* istanbul ignore next */ if (this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___post__ instanceof Function) this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___post__();
  }
}
const MyStruct = new Proxy($MyStruct, { apply(receiver, self, args) { return new $MyStruct(...args); } });
]])
    end)

    test("@Self struct Name", function()
      parser:parse("@Self struct MyStruct")
      assert(trans:next()):eq([[
const $MyStruct = class MyStruct {
  constructor(_) {
    /* istanbul ignore next */ if (_ == null) _ = {};
    /* istanbul ignore next */ if (this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___post__ instanceof Function) this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___post__();
  }
}
const MyStruct = new Proxy($MyStruct, { apply(receiver, self, args) { return new $MyStruct(...args); } });
const Self = MyStruct;
]])
    end)

    test("@emitter struct Name", function()
      parser:parse("@emitter struct MyStruct")
      assert(trans:next()):eq([[
const $MyStruct = class MyStruct extends dogma.EventEmitter {
  constructor(_) {
    super();
    /* istanbul ignore next */ if (_ == null) _ = {};
    /* istanbul ignore next */ if (this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___post__ instanceof Function) this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___post__();
  }
}
const MyStruct = new Proxy($MyStruct, { apply(receiver, self, args) { return new $MyStruct(...args); } });
]])
    end)

    test("@abstract struct Name", function()
      parser:parse("@abstract struct MyStruct")
      assert(trans:next()):eq([[
const $MyStruct = class MyStruct {
  constructor(_) {
    /* istanbul ignore next */ if (_ == null) _ = {};
    /* istanbul ignore next */ if (this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___post__ instanceof Function) this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___post__();
  }
}
const MyStruct = new Proxy($MyStruct, { apply(receiver, self, args) { /* istanbul ignore next */ throw("'MyStruct' is abstract."); } });
]])
    end)

    test("export struct Name", function()
      parser:parse("export struct MyStruct")
      assert(trans:next()):eq([[
const $MyStruct = class MyStruct {
  constructor(_) {
    /* istanbul ignore next */ if (_ == null) _ = {};
    /* istanbul ignore next */ if (this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___post__ instanceof Function) this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___post__();
  }
}
const MyStruct = new Proxy($MyStruct, { apply(receiver, self, args) { return new $MyStruct(...args); } });
module.exports = exports = MyStruct;
]])
    end)

    test("pub struct Name", function()
      parser:parse("pub struct MyStruct")
      assert(trans:next()):eq([[
const $MyStruct = class MyStruct {
  constructor(_) {
    /* istanbul ignore next */ if (_ == null) _ = {};
    /* istanbul ignore next */ if (this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___post__ instanceof Function) this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___post__();
  }
}
const MyStruct = new Proxy($MyStruct, { apply(receiver, self, args) { return new $MyStruct(...args); } });
module.exports['MyStruct'] = exports['MyStruct'] = MyStruct;
]])
    end)

    test("struct Name : Name", function()
      parser:parse("struct MyStruct : MyBase")
      assert(trans:next()):eq([[
const $MyStruct = class MyStruct extends MyBase {
  constructor(_) {
    super(_);
    /* istanbul ignore next */ if (_ == null) _ = {};
    /* istanbul ignore next */ if (this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___post__ instanceof Function) this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___post__();
  }
}
const MyStruct = new Proxy($MyStruct, { apply(receiver, self, args) { return new $MyStruct(...args); } });
]])
    end)

    test("var fields", function()
      parser:parse("struct MyStruct\n  pub var a\n  intl var b?\n  pub var c:num\n  pub var d?:num\n  pub var e:=123\n  pub var f:text=123\n  pub var g=123\n  pub var h=_.h\n  intl var i=_.i")
      assert(trans:next()):eq([[
const $MyStruct = class MyStruct {
  constructor(_) {
    /* istanbul ignore next */ if (_ == null) _ = {};
    expect('a', _['a'], null);
    Object.defineProperty(this, 'a', {value: coalesce(_['a'], null), writable: true, enumerable: true});
    Object.defineProperty(this, '_b', {value: coalesce(_['b'], null), writable: true, enumerable: false});
    expect('c', _['c'], num);
    Object.defineProperty(this, 'c', {value: coalesce(_['c'], null), writable: true, enumerable: true});
    if (_['d'] != null) expect('d', _['d'], num);
    Object.defineProperty(this, 'd', {value: coalesce(_['d'], null), writable: true, enumerable: true});
    if (_['e'] != null) expect('e', _['e'], num);
    Object.defineProperty(this, 'e', {value: coalesce(_['e'], 123), writable: true, enumerable: true});
    if (_['f'] != null) expect('f', _['f'], text);
    Object.defineProperty(this, 'f', {value: coalesce(_['f'], 123), writable: true, enumerable: true});
    Object.defineProperty(this, 'g', {value: coalesce(_['g'], 123), writable: true, enumerable: true});
    Object.defineProperty(this, 'h', {value: _.h, writable: true, enumerable: true});
    Object.defineProperty(this, '_i', {value: _.i, writable: true, enumerable: false});
    /* istanbul ignore next */ if (this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___post__ instanceof Function) this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___post__();
  }
}
const MyStruct = new Proxy($MyStruct, { apply(receiver, self, args) { return new $MyStruct(...args); } });
]])
    end)

    test("const fields", function()
      parser:parse("struct MyStruct\n  pub const a\n  intl const b=text\n  pub const c:text\n  pub const d:text=num\n  pub const e:=123")
      assert(trans:next()):eq([[
const $MyStruct = class MyStruct {
  constructor(_) {
    /* istanbul ignore next */ if (_ == null) _ = {};
    expect('a', _['a'], null);
    Object.defineProperty(this, 'a', {value: coalesce(_['a'], null), writable: false, enumerable: true});
    Object.defineProperty(this, '_b', {value: coalesce(_['b'], text), writable: false, enumerable: false});
    expect('c', _['c'], text);
    Object.defineProperty(this, 'c', {value: coalesce(_['c'], null), writable: false, enumerable: true});
    if (_['d'] != null) expect('d', _['d'], text);
    Object.defineProperty(this, 'd', {value: coalesce(_['d'], num), writable: false, enumerable: true});
    if (_['e'] != null) expect('e', _['e'], num);
    Object.defineProperty(this, 'e', {value: coalesce(_['e'], 123), writable: false, enumerable: true});
    /* istanbul ignore next */ if (this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___post__ instanceof Function) this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___post__();
  }
}
const MyStruct = new Proxy($MyStruct, { apply(receiver, self, args) { return new $MyStruct(...args); } });
]])
    end)

    test("private fields", function()
      parser:parse("struct MyStruct\n  pvt var a\n  pvt const b")
      assert(trans:next()):eq([[
const $MyStruct = class MyStruct {
  constructor(_) {
    /* istanbul ignore next */ if (_ == null) _ = {};
    expect('a', _['a'], null);
    Object.defineProperty(this, '_pvt_e6eace38db3ac6dfd7468b60ae7a70ae_a', {value: coalesce(_['a'], null), writable: true, enumerable: false});
    expect('b', _['b'], null);
    Object.defineProperty(this, '_pvt_e6eace38db3ac6dfd7468b60ae7a70ae_b', {value: coalesce(_['b'], null), writable: false, enumerable: false});
    /* istanbul ignore next */ if (this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___post__ instanceof Function) this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___post__();
  }
}
const MyStruct = new Proxy($MyStruct, { apply(receiver, self, args) { return new $MyStruct(...args); } });
]])
    end)

    test("struct methods", function()
      parser:parse("struct MyStruct\n  @post\n  pub fn method1()\n  @post\n  pvt proc method2()\n  @post\n  intl proc method3()\n  @abstract\n  pub proc method4()\n")
      assert(trans:next()):eq([[
const $MyStruct = class MyStruct {
  constructor(_) {
    /* istanbul ignore next */ if (_ == null) _ = {};
    /* istanbul ignore next */ if (this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___post__ instanceof Function) this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___post__();
  }
}
const MyStruct = new Proxy($MyStruct, { apply(receiver, self, args) { return new $MyStruct(...args); } });
MyStruct.prototype.method1 = function() { {} };
MyStruct.prototype._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___post__ = MyStruct.prototype.method1;
MyStruct.prototype._pvt_e6eace38db3ac6dfd7468b60ae7a70ae_method2 = function() { {} };
MyStruct.prototype._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___post__ = MyStruct.prototype._pvt_e6eace38db3ac6dfd7468b60ae7a70ae_method2;
MyStruct.prototype._method3 = function() { {} };
MyStruct.prototype._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___post__ = MyStruct.prototype._method3;
/* istanbul ignore next */ MyStruct.prototype.method4 = function() { abstract(); };
]])
    end)
  end):tags("struct")

  ----------
  -- type --
  ----------
  suite("type", function()
    test("@emitter type Name()", function()
      parser:parse("@emitter type Name()")
      assert(trans:next()):eq([[
const $Name = class Name extends dogma.EventEmitter {
  constructor() { super();{}  }
};
const Name = new Proxy($Name, { apply(receiver, self, args) { return new $Name(...args); } });
]])
    end)

    test("@todo type Name()", function()
      parser:parse("@todo type Coord2D()")
      assert(trans:next()):eq([[
const $Coord2D = class Coord2D {
  constructor() { todo(); }
};
const Coord2D = new Proxy($Coord2D, { apply(receiver, self, args) { return new $Coord2D(...args); } });
]])
    end)

    test("@Self type Name()", function()
      parser:parse("@Self type Coord2D()")
      assert(trans:next()):eq([[
const $Coord2D = class Coord2D {
  constructor() { {}  }
};
const Coord2D = new Proxy($Coord2D, { apply(receiver, self, args) { return new $Coord2D(...args); } });
const Self = Coord2D;
]])
    end)

    test("@component type Name", function()
      parser:parse("@component type Coord2D")
      assert(trans:next()):eq([[
const $Coord2D = class Coord2D {
  constructor($obj) { dogma.autobind(this, $Coord2D);/* istanbul ignore next */if ($obj == null) $obj = {};  }
};Object.defineProperty($Coord2D, '__bind__', {value: [], enumerable: false, writable: false});
const Coord2D = new Proxy($Coord2D, { apply(receiver, self, args) { return new $Coord2D(...args); } });
]])
    end)

    test("@Self type Name", function()
      parser:parse("@Self type Coord2D")
      assert(trans:next()):eq([[
const $Coord2D = class Coord2D {
  constructor($obj) { /* istanbul ignore next */if ($obj == null) $obj = {};  }
};
const Coord2D = new Proxy($Coord2D, { apply(receiver, self, args) { return new $Coord2D(...args); } });
const Self = Coord2D;
]])
    end)

    test("type Name", function()
      parser:parse("type Coord2D")
      assert(trans:next()):eq([[
const $Coord2D = class Coord2D {
  constructor($obj) { /* istanbul ignore next */if ($obj == null) $obj = {};  }
};
const Coord2D = new Proxy($Coord2D, { apply(receiver, self, args) { return new $Coord2D(...args); } });
]])
    end)

    test("@abstract export type Name", function()
      parser:parse("@abstract\nexport type Coord2D")
      assert(trans:next()):eq([[
const $Coord2D = class Coord2D {
  constructor($obj) { /* istanbul ignore next */if ($obj == null) $obj = {};  }
};
const Coord2D = new Proxy($Coord2D, { apply(receiver, self, args) { /* istanbul ignore next */ throw("'Coord2D' is abstract."); } });module.exports = exports = Coord2D;
]])
    end)

    test("type Name with fields", function()
      parser:parse('type Coord3D\n  const x\n  var y?\n  var z?:num\n  pvt const kind="3D"')
      assert(trans:next()):eq([[
const $Coord3D = class Coord3D {
  constructor($obj) { /* istanbul ignore next */if ($obj == null) $obj = {};/* istanbul ignore next */ dogma.paramExpected("x", $obj["x"], null);Object.defineProperty(this, "x", {value: $obj["x"], enum: true, writable: false});Object.defineProperty(this, "y", {value: $obj["y"], enum: true, writable: true});/* istanbul ignore next */if ($obj["z"] != null) dogma.paramExpected("z", $obj["z"], num);Object.defineProperty(this, "z", {value: $obj["z"], enum: true, writable: true});/* istanbul ignore next */if ($obj["kind"] != null) dogma.paramExpected("kind", $obj["kind"], null);Object.defineProperty(this, "_kind", {value: $obj["kind"]||"3D", writable: false});  }
};
const Coord3D = new Proxy($Coord3D, { apply(receiver, self, args) { return new $Coord3D(...args); } });
]])
    end)

    test("type Name()\\n Body", function()
      parser:parse("type Coord2D()\n .x = 0\n .y = 0")
      assert(trans:next()):eq([[
const $Coord2D = class Coord2D {
  constructor() { {(this.x=0);(this.y=0);}  }
};
const Coord2D = new Proxy($Coord2D, { apply(receiver, self, args) { return new $Coord2D(...args); } });
]])
    end)

    test("@component type Name()\\n Body", function()
      parser:parse("@component\ntype Coord2D()\n .x = 0\n .y = 0")
      assert(trans:next()):eq([[
const $Coord2D = class Coord2D {
  constructor() { dogma.autobind(this, $Coord2D);{(this.x=0);(this.y=0);}  }
};Object.defineProperty($Coord2D, '__bind__', {value: [], enumerable: false, writable: false});
const Coord2D = new Proxy($Coord2D, { apply(receiver, self, args) { return new $Coord2D(...args); } });
]])
    end)

    test("type Coord3D(.x, :y, !z)", function()
      parser:parse("type Coord2D(.x, :y, !z)")
      assert(trans:next()):eq([[
const $Coord2D = class Coord2D {
  constructor(x, y, z) { /* istanbul ignore next */ dogma.paramExpected("x", x, null);/* istanbul ignore next */ dogma.paramExpected("y", y, null);/* istanbul ignore next */ dogma.paramExpected("z", z, null);Object.defineProperty(this, "x", {value: x, enum: true, writable: true});Object.defineProperty(this, "_y", {value: y, writable: true});Object.defineProperty(this, "_pvt_e6eace38db3ac6dfd7468b60ae7a70ae_z", {value: z, writable: true});{}  }
};
const Coord2D = new Proxy($Coord2D, { apply(receiver, self, args) { return new $Coord2D(...args); } });
]])
    end)

    test("@component type Coord2D(.x, .y)", function()
      parser:parse("@component type Coord2D(.x, .y)")
      assert(trans:next()):eq([[
const $Coord2D = class Coord2D {
  constructor(x, y) { /* istanbul ignore next */ dogma.paramExpected("x", x, null);/* istanbul ignore next */ dogma.paramExpected("y", y, null);dogma.autobind(this, $Coord2D);Object.defineProperty(this, "x", {value: x, enum: true, writable: true});Object.defineProperty(this, "y", {value: y, enum: true, writable: true});{}  }
};Object.defineProperty($Coord2D, '__bind__', {value: [], enumerable: false, writable: false});
const Coord2D = new Proxy($Coord2D, { apply(receiver, self, args) { return new $Coord2D(...args); } });
]])
    end)

    test("type Name(:Name)", function()
      parser:parse("type Coord1D(:x)")
      assert(trans:next()):eq([[
const $Coord1D = class Coord1D {
  constructor(x) { /* istanbul ignore next */ dogma.paramExpected("x", x, null);Object.defineProperty(this, "_x", {value: x, writable: true});{}  }
};
const Coord1D = new Proxy($Coord1D, { apply(receiver, self, args) { return new $Coord1D(...args); } });
]])
    end)

    test("type Name() : Name", function()
      parser:parse("type Coord3D() : Coord2D")
      assert(trans:next()):eq([[
const $Coord3D = class Coord3D extends Coord2D {
  constructor() { {}  }
};
const Coord3D = new Proxy($Coord3D, { apply(receiver, self, args) { return new $Coord3D(...args); } });
]])
    end)

    test("type Name() : Name.Name()", function()
      parser:parse("type Coord3D() : Coord.D2()")
      assert(trans:next()):eq([[
const $Coord3D = class Coord3D extends Coord.D2 {
  constructor() { super();{}  }
};
const Coord3D = new Proxy($Coord3D, { apply(receiver, self, args) { return new $Coord3D(...args); } });
]])
    end)

    test("type Coord3D(x, y, .z) : Coord2D(x, y)", function()
      parser:parse("type Coord3D(x, y, .z) : Coord2D(x, y)")
      assert(trans:next()):eq([[
const $Coord3D = class Coord3D extends Coord2D {
  constructor(x, y, z) { /* istanbul ignore next */ dogma.paramExpected("x", x, null);/* istanbul ignore next */ dogma.paramExpected("y", y, null);/* istanbul ignore next */ dogma.paramExpected("z", z, null);super(x, y);Object.defineProperty(this, "z", {value: z, enum: true, writable: true});{}  }
};
const Coord3D = new Proxy($Coord3D, { apply(receiver, self, args) { return new $Coord3D(...args); } });
]])
    end)

    test("type Coord3D(x, y, .z) : Coord2D(...)", function()
      parser:parse("type Coord3D(x, y, .z) : Coord2D(...)")
      assert(trans:next()):eq([[
const $Coord3D = class Coord3D extends Coord2D {
  constructor(x, y, z) { /* istanbul ignore next */ dogma.paramExpected("x", x, null);/* istanbul ignore next */ dogma.paramExpected("y", y, null);/* istanbul ignore next */ dogma.paramExpected("z", z, null);super(...arguments);Object.defineProperty(this, "z", {value: z, enum: true, writable: true});{}  }
};
const Coord3D = new Proxy($Coord3D, { apply(receiver, self, args) { return new $Coord3D(...args); } });
]])
    end)

    test("@component type Coord3D(...) : Coord2D(...)", function()
      parser:parse("@component type Coord3D(...) : Coord2D(...)")
      assert(trans:next()):eq([[
const $Coord3D = class Coord3D extends Coord2D {
  constructor() { super(...arguments);dogma.autobind(this, $Coord3D);{}  }
};Object.defineProperty($Coord3D, '__bind__', {value: [], enumerable: false, writable: false});
const Coord3D = new Proxy($Coord3D, { apply(receiver, self, args) { return new $Coord3D(...args); } });
]])
    end)

    test("type Coord3D(...) : Coord2D(...)", function()
      parser:parse("type Coord3D(...) : Coord2D(...)")
      assert(trans:next()):eq([[
const $Coord3D = class Coord3D extends Coord2D {
  constructor() { super(...arguments);{}  }
};
const Coord3D = new Proxy($Coord3D, { apply(receiver, self, args) { return new $Coord3D(...args); } });
]])
    end)

    test("export type Coord2D()", function()
      parser:parse("export type Coord2D()")
      assert(trans:next()):eq([[
const $Coord2D = class Coord2D {
  constructor() { {}  }
};
const Coord2D = new Proxy($Coord2D, { apply(receiver, self, args) { return new $Coord2D(...args); } });module.exports = exports = Coord2D;
]])
    end)

    test("@abstract export type Coord2D()", function()
      parser:parse("@abstract\nexport type Coord2D()")
      assert(trans:next()):eq([[
const $Coord2D = class Coord2D {
  constructor() { {}  }
};
const Coord2D = new Proxy($Coord2D, { apply(receiver, self, args) { /* istanbul ignore next */ throw("'Coord2D' is abstract."); } });module.exports = exports = Coord2D;
]])
    end)

    suite("mixin", function()
      test("@mixable type Name()", function()
        parser:parse("@mixable type Coord2D(x, y)")
        assert(trans:next()):eq([[
const $Coord2D = class Coord2D {
  constructor() { this._constructor(...arguments); } _constructor(x, y) { /* istanbul ignore next */ dogma.paramExpected("x", x, null);/* istanbul ignore next */ dogma.paramExpected("y", y, null);{}  }
};
const Coord2D = new Proxy($Coord2D, { apply(receiver, self, args) { return new $Coord2D(...args); } });
]])
      end)

      test("type Name() :: Mixin()\\n Body", function()
        parser:parse("type A() :: B()\n .x = 0\n .y = 0")
        assert(trans:next()):eq([[
const $A = class A {
  constructor() { /* istanbul ignore next */if (B.prototype._constructor instanceof Function) B.prototype._constructor.bind(this)();{(this.x=0);(this.y=0);}  }
};dogma.mixin($A, B);
const A = new Proxy($A, { apply(receiver, self, args) { return new $A(...args); } });
]])
      end)

      test("type Name :: Mixin()", function()
        parser:parse("type A :: B()")
        assert(trans:next()):eq([[
const $A = class A {
  constructor($obj) { /* istanbul ignore next */if ($obj == null) $obj = {};/* istanbul ignore next */if (B.prototype._constructor instanceof Function) B.prototype._constructor.bind(this)();  }
};dogma.mixin($A, B);
const A = new Proxy($A, { apply(receiver, self, args) { return new $A(...args); } });
]])
      end)

      test("type Name() :: Mixin()", function()
        parser:parse("type A() :: B()")
        assert(trans:next()):eq([[
const $A = class A {
  constructor() { /* istanbul ignore next */if (B.prototype._constructor instanceof Function) B.prototype._constructor.bind(this)();{}  }
};dogma.mixin($A, B);
const A = new Proxy($A, { apply(receiver, self, args) { return new $A(...args); } });
]])
      end)

      test("type Name() :: Mixin(...)", function()
        parser:parse("type A() :: B(...)")
        assert(trans:next()):eq([[
const $A = class A {
  constructor() { /* istanbul ignore next */if (B.prototype._constructor instanceof Function) B.prototype._constructor.bind(this)(...arguments);{}  }
};dogma.mixin($A, B);
const A = new Proxy($A, { apply(receiver, self, args) { return new $A(...args); } });
]])
      end)

      test("type Name() :: Mixin(args)", function()
        parser:parse("type A() :: B(one, two)")
        assert(trans:next()):eq([[
const $A = class A {
  constructor() { /* istanbul ignore next */if (B.prototype._constructor instanceof Function) B.prototype._constructor.bind(this)(one, two);{}  }
};dogma.mixin($A, B);
const A = new Proxy($A, { apply(receiver, self, args) { return new $A(...args); } });
]])
      end)
    end):tags("mixin")
  end):tags("type")

  ----------
  -- proc --
  ----------
  suite("proc", function()
    suite("@input", function()
      test("procedure", function()
        parser:parse("@input\nproc f(x) x")
        assert(trans:next()):eq('function f(x) { print(...arguments);/* istanbul ignore next */ dogma.paramExpected("x", x, null);{x;} }\n')
      end)

      test("method", function()
        parser:parse("@input\npr Type.f(x) x")
        assert(trans:next()):eq('Type.prototype.f = function(x) { print(...arguments);/* istanbul ignore next */ dogma.paramExpected("x", x, null);{x;} };\n')
      end)
    end)

    suite("@bind", function()
      test("procedure", function()
        parser:parse("@bind\npr sum(x?, y?) x + y")
        assert(trans:next()):eq("function sum(x, y) { {(x+y);} }\n")
      end)

      test("method", function()
        parser:parse("@bind\nproc Calcul.sum(x?, y?) x + y")
        assert(trans:next()):eq("Calcul.__bind__.push(['sum', function(x, y) { {(x+y);} }]);\n")
      end)
    end)

    suite("standalone", function()
      test("@todo proc Name()", function()
        parser:parse("@todo\nproc myfn()")
        assert(trans:next()):eq("function myfn() { todo(); }\n")
      end)

      test("proc Name()", function()
        parser:parse("proc myfn()")
        assert(trans:next()):eq("function myfn() { {} }\n")
      end)

      suite("@alias", function()
        test("@alias proc Name()", function()
          parser:parse('@alias("myfn2") proc myfn()')
          assert(trans:next()):eq("function myfn() { {} } const myfn2 = myfn;\n")
        end)

        test("@alias pub proc Name()", function()
          parser:parse('@alias("myfn2") pub proc myfn()')
          assert(trans:next()):eq("function myfn() { {} }export {myfn}; const myfn2 = myfn;export {myfn2};\n")
        end)
      end)

      test("proc Name() Exp", function()
        parser:parse("proc myfn() 123")
        assert(trans:next()):eq("function myfn() { {123;} }\n")
      end)

      test("export proc Name()", function()
        parser:parse("export proc myfn()")
        assert(trans:next()):eq('function myfn() { {} }module.exports = exports = myfn;\n')
      end)

      test("pub proc Name()", function()
        parser:parse("pub proc myfn()")
        assert(trans:next()):eq('function myfn() { {} }export {myfn};\n')
      end)

      test("proc Name(params) Body", function()
        parser:parse("proc sum(x, y) x+y")
        assert(trans:next()):eq('function sum(x, y) { /* istanbul ignore next */ dogma.paramExpected("x", x, null);/* istanbul ignore next */ dogma.paramExpected("y", y, null);{(x+y);} }\n')
      end)

      test("async proc Name(params) Body", function()
        parser:parse("async proc sum(x, y) x+y")
        assert(trans:next()):eq('async function sum(x, y) { /* istanbul ignore next */ dogma.paramExpected("x", x, null);/* istanbul ignore next */ dogma.paramExpected("y", y, null);{(x+y);} }\n')
      end)
    end)

    suite("type", function()
      suite("static method", function()
        test("@static proc Name . Name() Body", function()
          parser:parse("@static\nproc MyType.method() 1+2")
          assert(trans:next()):eq("MyType.method = function() { {(1+2);} };\n")
        end)

        test("@static async proc Name . Name() Body", function()
          parser:parse("@static\nasync proc MyType.method() 1+2")
          assert(trans:next()):eq("MyType.method = async function() { {(1+2);} };\n")
        end)

        test("@static @todo proc Name . Name() Body", function()
          parser:parse("@static @todo\nproc MyType.method() 1+2")
          assert(trans:next()):eq("MyType.method = function() { todo(); };\n")
        end)
      end):tags("static")

      suite("@alias", function()
        test("@alias proc Name . Name ()", function()
          parser:parse('@alias("myfn2") proc MyType.myfn()')
          assert(trans:next()):eq('MyType.prototype.myfn2 = MyType.prototype.myfn = function() { {} };\n')
        end)
      end)

      test("proc Name . Name ()", function()
        parser:parse("proc MyType.myfn()")
        assert(trans:next()):eq('MyType.prototype.myfn = function() { {} };\n')
      end)

      test("async proc Name . Name ()", function()
        parser:parse("async proc MyType.myfn()")
        assert(trans:next()):eq('MyType.prototype.myfn = async function() { {} };\n')
      end)

      suite("abstract", function()
        test("pub", function()
          parser:parse("@abstract\nproc MyType.myfn()")
          assert(trans:next()):eq('/* istanbul ignore next */ MyType.prototype.myfn = function() { abstract(); };\n')
        end)

        test("intl", function()
          parser:parse("@abstract\nproc MyType:myfn()")
          assert(trans:next()):eq('/* istanbul ignore next */ MyType.prototype._myfn = function() { abstract(); };\n')
        end)

        test("pvt", function()
          parser:parse("@abstract\nproc MyType!myfn()")
          assert(function() trans:next() end):raises("anonymous code (2, 1): abstract method can't be private (!).")
        end)
      end):tags("abstract")

      test("@todo\\nproc Name . Name ()", function()
        parser:parse("@todo\nproc MyType.myfn()")
        assert(trans:next()):eq('MyType.prototype.myfn = function() { todo(); };\n')
      end)

      test("proc Name : Name ()", function()
        parser:parse("proc MyType:myfn()")
        assert(trans:next()):eq('MyType.prototype._myfn = function() { {} };\n')
      end)

      test("proc Name ! Name ()", function()
        parser:parse("proc MyType!myfn()")
        assert(trans:next()):eq('MyType.prototype._pvt_e6eace38db3ac6dfd7468b60ae7a70ae_myfn = function() { {} };\n')
      end)

      test("proc Name . Name (. Name)", function()
        parser:parse("proc MyType.myfn(.x)")
        assert(trans:next()):eq('MyType.prototype.myfn = function(x) { /* istanbul ignore next */ dogma.paramExpected("x", x, null);Object.defineProperty(this, "x", {value: x, enum: true, writable: true});{} };\n')
      end)

      test("proc Name . Name (. Name , . Name)", function()
        parser:parse("proc MyType.myfn(.x, .y)")
        assert(trans:next()):eq('MyType.prototype.myfn = function(x, y) { /* istanbul ignore next */ dogma.paramExpected("x", x, null);/* istanbul ignore next */ dogma.paramExpected("y", y, null);Object.defineProperty(this, "x", {value: x, enum: true, writable: true});Object.defineProperty(this, "y", {value: y, enum: true, writable: true});{} };\n')
      end)

      test("proc Name . Name (const . Name)", function()
        parser:parse("proc MyType.myfn(const .x)")
        assert(trans:next()):eq('MyType.prototype.myfn = function(x) { /* istanbul ignore next */ dogma.paramExpected("x", x, null);Object.defineProperty(this, "x", {value: x, enum: true, writable: false});{} };\n')
      end)

      test("proc Name . Name (const . Name)", function()
        parser:parse("proc MyType.myfn(const .x)")
        assert(trans:next()):eq('MyType.prototype.myfn = function(x) { /* istanbul ignore next */ dogma.paramExpected("x", x, null);Object.defineProperty(this, "x", {value: x, enum: true, writable: false});{} };\n')
      end)

      test("proc Name . Name (.x Name ?)", function()
        parser:parse("proc MyType.myfn(.x?)")
        assert(trans:next()):eq('MyType.prototype.myfn = function(x) { Object.defineProperty(this, "x", {value: x, enum: true, writable: true});{} };\n')
      end)
    end)
  end):tags("proc")

  --------
  -- fn --
  --------
  suite("fn", function()
    suite("@input", function()
      test("function", function()
        parser:parse("@input\nfn f(x) = x")
        assert(trans:next()):eq('function f(x) { print(...arguments);/* istanbul ignore next */ dogma.paramExpected("x", x, null);{return x;} }\n')
      end)

      test("method", function()
        parser:parse("@input\nfn Type.f(x) = x")
        assert(trans:next()):eq('Type.prototype.f = function(x) { print(...arguments);/* istanbul ignore next */ dogma.paramExpected("x", x, null);{return x;} };\n')
      end)
    end)

    suite("@bind", function()
      test("function", function()
        parser:parse("@bind\nfn sum(x?, y?) = x + y")
        assert(trans:next()):eq("function sum(x, y) { {return (x+y);} }\n")
      end)

      test("method", function()
        parser:parse("@bind\nfn Calcul.sum(x?, y?) = x + y")
        assert(trans:next()):eq("Calcul.__bind__.push(['sum', function(x, y) { {return (x+y);} }]);\n")
      end)
    end)

    suite("params", function()
      suite("unpack", function()
        suite("=>", function()
          test("fn(params=>[...])", function()
            parser:parse("fn sum(params=>[x,y]) = x+y")
            assert(trans:next()):eq('function sum(params) { /* istanbul ignore next */ dogma.paramExpected("params", params, list);let [x,y] = params;{return (x+y);} }\n')
          end)

          test("fn(params=>{...})", function()
            parser:parse("fn sum(params=>{x,y}) = x+y")
            assert(trans:next()):eq('function sum(params) { /* istanbul ignore next */ dogma.paramExpected("params", params, map);let {x,y} = params;{return (x+y);} }\n')
          end)

          test("fn(params:{...}=>{...})", function()
            parser:parse("fn sum(params:{x:num,y:num,z:num}=>{x,y}) = x+y")
            assert(trans:next()):eq('function sum(params) { /* istanbul ignore next */ dogma.paramExpectedToHave(\"params\", params, {x: {type: num, mandatory: true}, y: {type: num, mandatory: true}, z: {type: num, mandatory: true}}, true);let {x,y} = params;{return (x+y);} }\n')
          end)
        end)

        suite(":>", function()
          test("fn(params:>{...})", function()
            parser:parse("fn sum(params:>{x:num,y:num}) = x+y")
            assert(trans:next()):eq('function sum(params) { /* istanbul ignore next */ dogma.paramExpectedToHave(\"params\", params, {x: {type: num, mandatory: true}, y: {type: num, mandatory: true}}, true);let {x,y} = params;{return (x+y);} }\n')
          end)
        end)
      end)

      test("fn func(keyword)", function()
        parser:parse("fn sum(x, class) = x + class")
        assert(trans:next()):eq('function sum(x, class_) { /* istanbul ignore next */ dogma.paramExpected("x", x, null);/* istanbul ignore next */ dogma.paramExpected("class_", class_, null);{return (x+class_);} }\n')
      end)

      test("fn Name(Name)", function()
        parser:parse("fn myfn(x)")
        assert(trans:next()):eq('function myfn(x) { /* istanbul ignore next */ dogma.paramExpected("x", x, null);{} }\n')
      end)

      test("fn Name(Name:Name)", function()
        parser:parse("fn myfn(x:num)")
        assert(trans:next()):eq('function myfn(x) { /* istanbul ignore next */ dogma.paramExpected("x", x, num);{} }\n')
      end)

      test("fn Name(Name:(Name))", function()
        parser:parse("fn myfn(x:(num))")
        assert(trans:next()):eq('function myfn(x) { /* istanbul ignore next */ dogma.paramExpected("x", x, [num]);{} }\n')
      end)

      test("fn Name(Name:(Name, Name))", function()
        parser:parse("fn myfn(x:(text,list))")
        assert(trans:next()):eq('function myfn(x) { /* istanbul ignore next */ dogma.paramExpected("x", x, [text, list]);{} }\n')
      end)

      test("fn Name(Name:[])", function()
        parser:parse("fn myfn(x:[])")
        assert(trans:next()):eq("function myfn(x) { x = list(x);{} }\n")
      end)

      test("fn Name(Name?:[])", function()
        parser:parse("fn myfn(x:[])")
        assert(trans:next()):eq("function myfn(x) { x = list(x);{} }\n")
      end)

      test("fn Name(Name:[Name])", function()
        parser:parse("fn myfn(x:[num])")
        assert(trans:next()):eq('function myfn(x) { /* istanbul ignore next */ dogma.paramExpected("x", x, [list, num]);x = list(x);{} }\n')
      end)

      test("fn Name(Name:[Name,Name])", function()
        parser:parse("fn myfn(x:[text,num])")
        assert(trans:next()):eq('function myfn(x) { /* istanbul ignore next */ dogma.paramExpected("x", x, [list, text, num]);x = list(x);{} }\n')
      end)

      test("fn Name(const Name)", function()
        parser:parse("fn myfn(x)")
        assert(trans:next()):eq('function myfn(x) { /* istanbul ignore next */ dogma.paramExpected("x", x, null);{} }\n')
      end)

      test("fn Name(Name?)", function()
        parser:parse("fn myfn(x?)")
        assert(trans:next()):eq('function myfn(x) { {} }\n')
      end)

      test("fn Name(Name?:Name)", function()
        parser:parse("fn myfn(x?:num)")
        assert(trans:next()):eq('function myfn(x) { /* istanbul ignore next */ dogma.paramExpectedToBe("x", x, num);{} }\n')
      end)

      test("fn Name(Name?:(Name))", function()
        parser:parse("fn myfn(x?:(num))")
        assert(trans:next()):eq('function myfn(x) { /* istanbul ignore next */ dogma.paramExpectedToBe("x", x, [num]);{} }\n')
      end)

      test("fn Name(Name?:(Name, Name))", function()
        parser:parse("fn myfn(x?:(text,list))")
        assert(trans:next()):eq('function myfn(x) { /* istanbul ignore next */ dogma.paramExpectedToBe("x", x, [text, list]);{} }\n')
      end)

      test("fn Name(Name?:[Name])", function()
        parser:parse("fn myfn(x?:[num])")
        assert(trans:next()):eq('function myfn(x) { /* istanbul ignore next */ dogma.paramExpectedToBe("x", x, [list, num]);if (x != null) x = list(x);{} }\n')
      end)

      test("fn Name(Name?:[Name,Name])", function()
        parser:parse("fn myfn(x?:[text,num])")
        assert(trans:next()):eq('function myfn(x) { /* istanbul ignore next */ dogma.paramExpectedToBe("x", x, [list, text, num]);if (x != null) x = list(x);{} }\n')
      end)

      test("fn Name(...Name)", function()
        parser:parse("fn myfn(...x)")
        assert(trans:next()):eq('function myfn(...x) { {} }\n')
      end)

      test("fn Name(Name=Name)", function()
        parser:parse("fn myfn(x=123)")
        assert(trans:next()):eq('function myfn(x = 123) { {} }\n')
      end)

      test("fn Name(Name:{})", function()
        parser:parse("fn sum(vals:{}) = vals.x + vals.y")
        assert(trans:next()):eq('function sum(vals) { /* istanbul ignore next */ dogma.paramExpectedToHave("vals", vals, {}, true);{return (vals.x+vals.y);} }\n')
      end)

      test("fn Name(Name:{p,q:type,r?:type,s:(type1,type2),t:[type1,type2]},u?:[type1,type2])", function()
        parser:parse("fn sum(vals:{p, q:num, r?:num, s:(text,list), t:[text,num], u?:[text,num]}) = vals.x + vals.y")
        assert(trans:next()):eq('function sum(vals) { /* istanbul ignore next */ dogma.paramExpectedToHave("vals", vals, {p: {type: any, mandatory: true}, q: {type: num, mandatory: true}, r: {type: num, mandatory: false}, s: {type: [text, list], mandatory: true}, t: {type: [list, text, num], mandatory: true}, u: {type: [list, text, num], mandatory: false}}, true);vals.t = list(vals.t);if (vals.u != null) vals.u = list(vals.u);{return (vals.x+vals.y);} }\n')
      end)

      test("fn Name(Name?:{p, q:type, r?:type})", function()
        parser:parse("fn sum(vals?:{x, y:num, z?:num}) = vals.x + vals.y")
        assert(trans:next()):eq('function sum(vals) { /* istanbul ignore next */ dogma.paramExpectedToHave("vals", vals, {x: {type: any, mandatory: true}, y: {type: num, mandatory: true}, z: {type: num, mandatory: false}}, false);{return (vals.x+vals.y);} }\n')
      end)
    end)

    suite("return var", function()
      test("fn Name() -> self Body", function()
        parser:parse("fn myfn() -> self\n 1+2")
        assert(trans:next()):eq("function myfn() { {(1+2);} return this; }\n")
      end)

      test("fn Name() -> Name Body", function()
        parser:parse("fn myfn() -> x\n 1+2")
        assert(trans:next()):eq("function myfn() { let x;{(1+2);} return x; }\n")
      end)

      test("fn Name(param) -> Name Body - return variable not existing", function()
        parser:parse("fn myfn(a?, b?) -> x\n x = a + b")
        assert(trans:next()):eq("function myfn(a, b) { let x;{(x=(a+b));} return x; }\n")
      end)

      test("fn Name(param) -> Name Body - return variable existing", function()
        parser:parse("fn myfn(a?, b?, x?) -> x\n x = a + b")
        assert(trans:next()):eq("function myfn(a, b, x) { {(x=(a+b));} return x; }\n")
      end)

      test("fn Name() -> res:map", function()
        parser:parse("fn sum(x?, y?) -> res:map\n res.result = x + y")
        assert(trans:next()):eq("function sum(x, y) { let res = {};{(res.result=(x+y));} return res; }\n")
      end)

      test("fn Name() -> res:list", function()
        parser:parse("fn sum(x?, y?) -> res:list\n res.push(x + y)")
        assert(trans:next()):eq("function sum(x, y) { let res = [];{res.push((x+y));} return res; }\n")
      end)

      test("fn Name() -> res:set", function()
        parser:parse("fn sum(x?, y?) -> res:set\n res.add(x + y)")
        assert(trans:next()):eq("function sum(x, y) { let res = set();{res.add((x+y));} return res; }\n")
      end)

      test("fn Name() -> res:bool", function()
        parser:parse("fn sum(x?, y?) -> res:bool")
        assert(trans:next()):eq("function sum(x, y) { let res = false;{} return res; }\n")
      end)

      test("fn Name() -> res:text", function()
        parser:parse("fn sum(x?, y?) -> res:text")
        assert(trans:next()):eq('function sum(x, y) { let res = "";{} return res; }\n')
      end)

      test("fn Name() -> res:text", function()
        parser:parse("fn sum(x?, y?) -> res:num")
        assert(trans:next()):eq('function sum(x, y) { let res = 0;{} return res; }\n')
      end)
    end)

    suite("standalone", function()
      suite("@alias", function()
        test("pub", function()
          parser:parse('@alias("f2")\npub fn f1(x) = x')
          assert(trans:next()):eq('function f1(x) { /* istanbul ignore next */ dogma.paramExpected("x", x, null);{return x;} }export {f1}; const f2 = f1;export {f2};\n')
        end)

        test("not pub", function()
          parser:parse('@alias("f2")\nfn f1(x) = x')
          assert(trans:next()):eq('function f1(x) { /* istanbul ignore next */ dogma.paramExpected("x", x, null);{return x;} } const f2 = f1;\n')
        end)
      end)

      test("@iter fn Name() Body", function()
        parser:parse("@iter\nfn myfn()\n yield 1+2")
        assert(trans:next()):eq("function* myfn() { {yield (1+2);} }\n")
      end):tags("yield")

      test("@todo fn Name()", function()
        parser:parse("@todo\nfn myfn()")
        assert(trans:next()):eq("function myfn() { todo(); }\n")
      end)

      test("fn Name()", function()
        parser:parse("fn myfn()")
        assert(trans:next()):eq("function myfn() { {} }\n")
      end)

      test("fn Name() = Exp", function()
        parser:parse("fn myfn() = 123")
        assert(trans:next()):eq("function myfn() { {return 123;} }\n")
      end)

      test("export fn Name()", function()
        parser:parse("export fn myfn()")
        assert(trans:next()):eq('function myfn() { {} }module.exports = exports = myfn;\n')
      end)

      test("pub fn Name()", function()
        parser:parse("pub fn myfn()")
        assert(trans:next()):eq('function myfn() { {} }export {myfn};\n')
      end)

      test("fn Name(params) Body", function()
        parser:parse("fn sum(x, y)\n return x+y")
        assert(trans:next()):eq('function sum(x, y) { /* istanbul ignore next */ dogma.paramExpected("x", x, null);/* istanbul ignore next */ dogma.paramExpected("y", y, null);{return (x+y);} }\n')
      end)

      test("async fn Name(params) Body", function()
        parser:parse("async fn sum(x, y)\n return x+y")
        assert(trans:next()):eq('async function sum(x, y) { /* istanbul ignore next */ dogma.paramExpected("x", x, null);/* istanbul ignore next */ dogma.paramExpected("y", y, null);{return (x+y);} }\n')
      end)

      test("async fn Name(params) -> var Body", function()
        parser:parse("async fn sum(x, y) -> r\n r = x+y")
        assert(trans:next()):eq('async function sum(x, y) { let r;/* istanbul ignore next */ dogma.paramExpected("x", x, null);/* istanbul ignore next */ dogma.paramExpected("y", y, null);{(r=(x+y));} return r; }\n')
      end)
    end)

    suite("type", function()
      suite("iter", function()
        test("@iter fn Name . iter () Body", function()
          parser:parse("@iter\nfn MyType.iter()\n yield 1+2")
          assert(trans:next()):eq("MyType.prototype[Symbol.iterator] = function*() {yield (1+2);};\n")
        end)

        test("@iter @todo fn Name . iter () Body", function()
          parser:parse("@iter @todo\nfn MyType.iter()\n yield 1+2")
          assert(trans:next()):eq("MyType.prototype[Symbol.iterator] = function*() { todo(); };\n")
        end)

        test("@iter fn Name . Name () Body", function()
          parser:parse("@iter\nfn MyType.iter2()\n yield 1+2")
          assert(trans:next()):eq("MyType.prototype.iter2 = function*() {yield (1+2);};\n")
        end)

        test("@iter @todo fn Name . Name () Body", function()
          parser:parse("@iter @todo\nfn MyType.iter2()\n yield 1+2")
          assert(trans:next()):eq("MyType.prototype.iter2 = function*() { todo(); };\n")
        end)
      end):tags("yield")

      suite("property", function()
        test("@prop fn Name . Name () = Exp", function()
          parser:parse("@prop\nfn MyType.prop() = 1+2")
          assert(trans:next()):eq('Object.defineProperty(MyType.prototype, "prop", {enum: true, get: function() { {return (1+2);} }});\n')
        end)

        test("@prop @todo fn Name . Name () = Exp", function()
          parser:parse("@prop @todo\nfn MyType.prop() = 1+2")
          assert(trans:next()):eq('Object.defineProperty(MyType.prototype, "prop", {enum: true, get: function() { todo(); }});\n')
        end)

        test("@prop fn Name . Name () -> Var Body", function()
          parser:parse("@prop\nfn MyType.prop() -> resp\n  resp = 1+2")
          assert(trans:next()):eq('Object.defineProperty(MyType.prototype, "prop", {enum: true, get: function() { let resp;{(resp=(1+2));} return resp; }});\n')
        end)

        test("@prop fn Name : Name () = Exp", function()
          parser:parse("@prop\nfn MyType:prop() = 1+2")
          assert(trans:next()):eq('Object.defineProperty(MyType.prototype, "_prop", {enum: false, get: function() { {return (1+2);} }});\n')
        end)

        test("@prop fn Name ! Name () = Exp", function()
          parser:parse("@prop\nfn MyType!prop() = 1+2")
          assert(trans:next()):eq('Object.defineProperty(MyType.prototype, "_pvt_e6eace38db3ac6dfd7468b60ae7a70ae_prop", {enum: false, get: function() { {return (1+2);} }});\n')
        end)

        test("@abstract @prop fn Name.Name()", function()
          parser:parse("@abstract @prop\nfn MyType.prop()")
          assert(trans:next()):eq('Object.defineProperty(MyType.prototype, "prop", {enum: true, get: function() { /* istanbul ignore next */abstract(); }});\n')
        end)
      end):tags("prop")

      suite("static method", function()
        test("@static fn Name . Name() -> res Body", function()
          parser:parse("@static\nfn MyType.method() -> res\n  res = 1+2")
          assert(trans:next()):eq("MyType.method = function() { let res;{(res=(1+2));} return res; };\n")
        end)

        test('@static @alias("alias") fn Name . Name() -> res Body', function()
          parser:parse('@static @alias("func")\nfn MyType.method() -> res\n  res = 1+2')
          assert(trans:next()):eq("MyType.func = MyType.method = function() { let res;{(res=(1+2));} return res; };\n")
        end)

        test("@static async fn Name . Name() Body", function()
          parser:parse("@static\nasync fn MyType.method() = 1+2")
          assert(trans:next()):eq("MyType.method = async function() { {return (1+2);} };\n")
        end)

        test("@static @todo fn Name . Name() Body", function()
          parser:parse("@static @todo\nfn MyType.method() = 1+2")
          assert(trans:next()):eq("MyType.method = function() { todo(); };\n")
        end)
      end):tags("static")

      suite("abstract", function()
        test("pub", function()
          parser:parse("@abstract\nfn MyType.myfn()")
          assert(trans:next()):eq('/* istanbul ignore next */ MyType.prototype.myfn = function() { abstract(); };\n')
        end)

        test("intl", function()
          parser:parse("@abstract\nfn MyType:myfn()")
          assert(trans:next()):eq('/* istanbul ignore next */ MyType.prototype._myfn = function() { abstract(); };\n')
        end)

        test("pvt", function()
          parser:parse("@abstract\nfn MyType!myfn()")
          assert(function() trans:next() end):raises("anonymous code (2, 1): abstract method can't be private (!).")
        end)
      end):tags("abstract")

      suite("instance", function()
        test("@alias", function()
          parser:parse('@alias("myfn2") fn MyType.myfn()')
          assert(trans:next()):eq('MyType.prototype.myfn2 = MyType.prototype.myfn = function() { {} };\n')
        end)

        test("async fn Name . Name ()", function()
          parser:parse("async fn MyType.myfn()")
          assert(trans:next()):eq('MyType.prototype.myfn = async function() { {} };\n')
        end)

        test('@alias("alias") async fn Name . Name ()', function()
          parser:parse('@alias("mymeth") async fn MyType.myfn()')
          assert(trans:next()):eq('MyType.prototype.mymeth = MyType.prototype.myfn = async function() { {} };\n')
        end)

        test("fn Name . Name ()", function()
          parser:parse("fn MyType.myfn()")
          assert(trans:next()):eq('MyType.prototype.myfn = function() { {} };\n')
        end)

        test('@alias("alias") fn Name . Name ()', function()
          parser:parse('@alias("mymeth") fn MyType.myfn()')
          assert(trans:next()):eq('MyType.prototype.mymeth = MyType.prototype.myfn = function() { {} };\n')
        end)

        test("@todo\\nfn Name . Name ()", function()
          parser:parse("@todo\nfn MyType.myfn()")
          assert(trans:next()):eq('MyType.prototype.myfn = function() { todo(); };\n')
        end)

        test("fn Name : Name ()", function()
          parser:parse("fn MyType:myfn()")
          assert(trans:next()):eq('MyType.prototype._myfn = function() { {} };\n')
        end)

        test("fn Name ! Name ()", function()
          parser:parse("fn MyType!myfn()")
          assert(trans:next()):eq('MyType.prototype._pvt_e6eace38db3ac6dfd7468b60ae7a70ae_myfn = function() { {} };\n')
        end)

        test("fn Name . Name (. Name)", function()
          parser:parse("fn MyType.myfn(.x)")
          assert(trans:next()):eq('MyType.prototype.myfn = function(x) { /* istanbul ignore next */ dogma.paramExpected("x", x, null);Object.defineProperty(this, "x", {value: x, enum: true, writable: true});{} };\n')
        end)

        test("fn Name . Name (. Name , . Name)", function()
          parser:parse("fn MyType.myfn(.x, .y)")
          assert(trans:next()):eq('MyType.prototype.myfn = function(x, y) { /* istanbul ignore next */ dogma.paramExpected("x", x, null);/* istanbul ignore next */ dogma.paramExpected("y", y, null);Object.defineProperty(this, "x", {value: x, enum: true, writable: true});Object.defineProperty(this, "y", {value: y, enum: true, writable: true});{} };\n')
        end)

        test("fn Name . Name (const . Name)", function()
          parser:parse("fn MyType.myfn(const .x)")
          assert(trans:next()):eq('MyType.prototype.myfn = function(x) { /* istanbul ignore next */ dogma.paramExpected("x", x, null);Object.defineProperty(this, "x", {value: x, enum: true, writable: false});{} };\n')
        end)

        test("fn Name . Name (const . Name)", function()
          parser:parse("fn MyType.myfn(const .x)")
          assert(trans:next()):eq('MyType.prototype.myfn = function(x) { /* istanbul ignore next */ dogma.paramExpected("x", x, null);Object.defineProperty(this, "x", {value: x, enum: true, writable: false});{} };\n')
        end)

        test("fn Name . Name (.x Name ?)", function()
          parser:parse("fn MyType.myfn(.x?)")
          assert(trans:next()):eq('MyType.prototype.myfn = function(x) { Object.defineProperty(this, "x", {value: x, enum: true, writable: true});{} };\n')
        end)

        test("fn Name . Name() -> self\\n Body", function()
          parser:parse("fn MyType.method() -> self\n 1+2")
          assert(trans:next()):eq("MyType.prototype.method = function() { {(1+2);} return this; };\n")
        end)

        test("fn Name . Name() -> self\\n Body", function()
          parser:parse("fn MyType.method() -> self\n 1+2")
          assert(trans:next()):eq("MyType.prototype.method = function() { {(1+2);} return this; };\n")
        end)
      end):tags("instance")
    end)
  end):tags("fn")

  --------
  -- if --
  --------
  suite("if", function()
    test("#[!cov ignore] if Exp then...", function()
      parser:parse("#[!cov ignore] if true then x+y")
      assert(trans:next()):eq("/*istanbul ignore next*/if (true) {(x+y);}\n")
    end)

    test("#[!cov ignore else] if Exp then...", function()
      parser:parse("#[!cov ignore else] if true then x+y else y+x")
      assert(trans:next()):eq("/*istanbul ignore else*/if (true) {(x+y);} else {(y+x);}\n")
    end)

    test("#[!cov ignore] if const x = Exp then...", function()
      parser:parse("#[!cov ignore] if const x = true then x+y")
      assert(trans:next()):eq("{/*istanbul ignore next*/const x=true; /*istanbul ignore next*/if (x) {(x+y);}}\n")
    end)

    test("#[!cov ignore else] if const x = Exp then...", function()
      parser:parse("#[!cov ignore else] if const x = true then x+y")
      assert(trans:next()):eq("{const x=true; /*istanbul ignore else*/if (x) {(x+y);}}\n")
    end)

    test("#[!cov ignore] if x; y then...", function()
      parser:parse("#[!cov ignore] if x; y then x+y")
      assert(trans:next()):eq("/*istanbul ignore next*/x; /*istanbul ignore next*/if (y) {(x+y);}\n")
    end)

    test("#[!cov ignore else] if x; y then...", function()
      parser:parse("#[!cov ignore else] if x; y then x+y")
      assert(trans:next()):eq("x; /*istanbul ignore else*/if (y) {(x+y);}\n")
    end)

    test("if Exp then Exp", function()
      parser:parse("if true then x+y")
      assert(trans:next()):eq("if (true) {(x+y);}\n")
    end)

    test("if Exp; Exp then Exp", function()
      parser:parse("if x+=1; x==8 then x+y")
      assert(trans:next()):eq("(x+=1); if ((x==8)) {(x+y);}\n")
    end)

    test("if var Exp; Exp then Exp", function()
      parser:parse("if var x=1; x then x+y")
      assert(trans:next()):eq("{let x=1; if (x) {(x+y);}}\n")
    end)

    test("if Name := Name; Exp then Exp", function()
      parser:parse("if x:=1; x then x+y")
      assert(trans:next()):eq("{let x=1; if (x) {(x+y);}}\n")
    end)

    test("if Name := Name then Exp", function()
      parser:parse("if x:=1; x then x+y")
      assert(trans:next()):eq("{let x=1; if (x) {(x+y);}}\n")
    end)

    test("if [Name, Name] = Exp; Exp then Exp", function()
      parser:parse("if [x, y] = arr; x then x+y")
      assert(trans:next()):eq("([x, y]=arr); if (x) {(x+y);}\n")
    end)

    test("if [Name, Name] := Exp; Exp then Exp", function()
      parser:parse("if [x, y] := arr; x then x+y")
      assert(trans:next()):eq("{let [x, y] = arr; if (x) {(x+y);}}\n")
    end)

    test("#[!cov ignore next] if [Name, Name] := Exp; Exp then Exp", function()
      parser:parse("#[!cov ignore next] if [x, y] := arr; x then x+y")
      assert(trans:next()):eq("{/*istanbul ignore next*/let [x, y] = arr; /*istanbul ignore next*/if (x) {(x+y);}}\n")
    end)

    test("if [Name, Name] ::= Exp; Exp then Exp", function()
      parser:parse("if [x, y] ::= arr; x then x+y")
      assert(trans:next()):eq("{const [x, y] = arr; if (x) {(x+y);}}\n")
    end)

    test("if Name ::= Exp; Exp then Exp", function()
      parser:parse("if x::=1; x then x+y")
      assert(trans:next()):eq("{const x=1; if (x) {(x+y);}}\n")
    end)

    test("if const Exp; Exp then Exp", function()
      parser:parse("if const x=1; x then x+y")
      assert(trans:next()):eq("{const x=1; if (x) {(x+y);}}\n")
    end)

    test("if Exp then Sent", function()
      parser:parse("if true then return x+y")
      assert(trans:next()):eq("if (true) {return (x+y);}\n")
    end)

    test("if Exp then Exp else Exp", function()
      parser:parse("if true then x+y else a+b")
      assert(trans:next()):eq("if (true) {(x+y);} else {(a+b);}\n")
    end)

    test("if Exp then Exp else Sent", function()
      parser:parse("if true then x+y else return a+b")
      assert(trans:next()):eq("if (true) {(x+y);} else {return (a+b);}\n")
    end)

    test("if with else if", function()
      parser:parse("if true then\n x+y\nelse if false then\n a+b")
      assert(trans:next()):eq("if (true) {(x+y);} else if (false) {(a+b);}\n")
    end)

    test("if with else if and else", function()
      parser:parse("if x then\n x+y\nelse if y then\n a+b\nelse if z then\n 1+2\nelse\n 3+4")
      assert(trans:next()):eq("if (x) {(x+y);} else if (y) {(a+b);} else if (z) {(1+2);} else {(3+4);}\n")
    end)
  end):tags("if")

  ---------
  -- pub --
  ---------
  suite("pub", function()
    test("pub Item", function()
      parser:parse("pub Item")
      assert(trans:next()):eq("export {Item};\n")
    end)

    test('pub "./Item"', function()
      parser:parse('pub "./Item"')
      assert(trans:next()):eq('const Item = dogma.use(require("./Item"));export {Item};\n')
    end)

    test('pub Item, "./Item"', function()
      parser:parse('pub Item1, "./Item2"')
      assert(trans:next()):eq('export {Item1};const Item2 = dogma.use(require("./Item2"));export {Item2};\n')
    end)
  end):tags("pub")

  ------------
  -- export --
  ------------
  suite("export", function()
    test("export Exp", function()
      parser:parse("export 1+2")
      assert(trans:next()):eq("module.exports = exports = (1+2);\n")
    end)
  end):tags("export")

  ----------
  -- with --
  ----------
  suite("with", function()
    test("wihtout coverage directive", function()
      parser:parse("with 1+2\n  if 1 then x+1\n  if 2 then y+1\n  else z+1")
      assert(trans:next()):eq("{const _ = (1+2); switch (_) { case 1: {(x+1);}; /* istanbul ignore next */break; case 2: {(y+1);}; /* istanbul ignore next */break; default: {(z+1);}}}\n")
    end)

    test("with coverage directive: else", function()
      parser:parse("#!coverage ignore else\nwith 1+2\n  if 1 then x+1\n  if 2 then y+1\n  else z+1")
      assert(trans:next()):eq("{const _ = (1+2); switch (_) { case 1: {(x+1);}; /* istanbul ignore next */break; case 2: {(y+1);}; /* istanbul ignore next */break; /*istanbul ignore next*/ default: {(z+1);}}}\n")
    end)

    test("with coverage directive: next", function()
      parser:parse("#!coverage ignore next\nwith 1+2\n  if 1 then x+1\n  if 2 then y+1\n  else z+1")
      assert(trans:next()):eq("{const _ = (1+2);/*istanbul ignore next*/ switch (_) { case 1: {(x+1);}; /* istanbul ignore next */break; case 2: {(y+1);}; /* istanbul ignore next */break; default: {(z+1);}}}\n")
    end)
  end):tags("with")
end):tags("stmt")
